
love.window.setTitle("Piecemeal Vertex Update Test")

--[[
	This program initializes a mock tilemap, a "texture atlas" (which is really just a gradient) plus quads, and a set of SpriteBatch + Attribute Meshes. On every frame, it updates one vertex attribute per layer with mesh:setVertex(). Sweep the mouse horizontally across the window to change which index is modified (left is lower.)

	The map defaults to 256x256 with 1 layer. There is one SpriteBatch+Mesh pair per layer*. Map dimensions and layer count can be changed with CLI args (see love.load()) or by modifying the initial variable assignments. All tilemap layers use the same Lua array for data: only the SpriteBatches and Meshes are unique.

	*(I had a version of this with multiple Batch+Mesh pairs per layer, but the partitioning made it harder to read. Hopefully I haven't broken anything while removing it.)

	The use-case which prompted this test was to incrementally modify per-tile attributes in a tilemap, cycling through each tile over the span of multiple frames.
	
	Something close to my use-case would be:
		love . mapsq 256 layers 4
	
	LÖVE version: 11.3
	
	My PC's specs, FWIW:
		CPU: Intel® Core™ i5-4430 CPU @ 3.00GHz × 4 
		GPU: NVIDIA Corporation GM107 [GeForce GTX 750] / GPU Driver Version: 495.44 CUDA Version: 11.5
		OS: Fedora Linux 35 (Workstation Edition) 64-bit
		RAM: 24 GB
--]]


-- Defaults which are affected by CLI args.
local map_sq = 256
local layers_per_map = 1
local jit_enabled = true
local zoom = 1.0 -- 1.0 == 800x600. Updated at run-time whenever the window is resized.


-- Set in love.load()
local map_cols, map_rows
local n_tiles, n_verts, vert_fmt
local map, map_cols, map_rows, tile_w, tile_h, quads, atlas, atlas_cols, atlas_rows
local sets -- (sets of per-layer SpriteBatches+Meshes)
local pseudo -- randomGenerator. Uses a fixed seed to get the same map layout on every run.


-- Demo state
local active_set = 1
local do_update_vertex = true
local use_shader = true
local vertex_to_update = 1
local render_map = true
local do_vsync = true
local show_atlas = false

local dbg_print_vertex_update_info = false -- Debug: print vertex write info to console

-- Timing
local time_set_vertex = 0
local time_draw_batch = 0


-- Custom vertex format table
local vert_fmt = {
	{"foobar", "float", 4},
}


-- A table to use for mesh:setVertex()
local t_vert = {0.0, 0.0, 0.0, 0.0}

-- ByteData to do the equivalent with FFI
local ffi = require("ffi")
local vert_data = love.data.newByteData(4*4)


-- Helper / init functions
-- Makes a new set of SpriteBatches and attribute Meshes.
local function newBatchSet(tex, layer_count, tile_count, vertex_count, vertex_format, usage)
	
	local layers = {}
	for i = 1, layer_count do
		local layer = {}
		layer.usage = usage
		
		layer.batch = love.graphics.newSpriteBatch(tex, tile_count, usage)
		layer.mesh = love.graphics.newMesh(vert_fmt, vertex_count, "points", usage)
		
		layer.batch:attachAttribute("foobar", layer.mesh)
		
		layers[i] = layer
	end
	
	return layers
end

-- Converts 2D coord to 1D index
local function mapFlat(x, y)
	return 1 + y*map_cols + x
end

-- Applies map state to a SpriteBatch.
local function initBatch(layers)
	
	-- Clear the SpriteBatch first
	for i = 1, #layers do
		layers[i].batch:clear()
	end
	
	for l = 1, #layers do
		for y = 0, map_rows-1 do
			for x = 0, map_cols-1 do
				local layer = layers[l]
				
				local map_i = mapFlat(x, y)
				
				layer.batch:add(quads[map[map_i]], x*tile_w, y*tile_h, 0, 1, 1, 0, 0, 0, 0)
				
				-- The mesh defaults (all 0's) would be fine to start with. Setting something makes it
				-- clear if the shader is currently running, though.
				local mesh = layer.mesh
				local vert_id = (map_i - 1)*4 + 1

				for vert_offset = 0, 3 do
					t_vert[1] = 0.0
					t_vert[2] = 0.1
					t_vert[3] = 0.5
					t_vert[4] = 1.0
					
					mesh:setVertex(vert_id + vert_offset, t_vert)
				end
			end
		end
	end
end


-- A shader which uses attributes as 'varying' state
local mock_shader = love.graphics.newShader([[//#pragma language glsl3
#ifdef VERTEX

attribute vec4 foobar;
varying vec4 v_foobar;

vec4 position(mat4 transform_projection, vec4 vertex_position)
{
	v_foobar = foobar;
	return transform_projection * vertex_position;
}

#endif

#ifdef PIXEL

varying vec4 v_foobar;

vec4 effect(vec4 color, Image tex, vec2 texture_coords, vec2 screen_coords)
{
	vec4 texturecolor = Texel(tex, texture_coords);
	
	// Some mock activity involving 'foobar'. My actual use case is indexed palette simulation and a dithered dissolve mask.
	color += v_foobar;
	
	// Debug: test if shader is being applied at all
	//texturecolor.r = 0.0;
	return texturecolor * color;
}

#endif
]]
)


function love.load(arguments)
	
	-- Parse args
	local i = 1
	while i <= #arguments do
		local action = arguments[i]

		if action == "jitoff" then
			jit_enabled = false
			jit.off()
			
		elseif action == "vsync_off" then
			-- Can also be toggled at runtime
			love.window.setVSync(0)
			do_vsync = false
			
		elseif action == "mapsq" then
			-- Warning: High map_sq values get unstable very very quickly. 2048 is OK on my system (in the sense that it doesn't immediately run out of memory and crash.)
			-- In my actual use-case, maps are almost always 256x256 or smaller.
			local arg2 = arguments[i + 1]
			local in_sq = math.floor(arg2) -- string->number coercion
			map_sq = in_sq
			i = i + 1
			
		elseif action == "layers" then
			-- Too many layers can also be a problem. My use-case would have at most 12 layers at a time.
			local arg2 = arguments[i + 1]
			local in_layers = math.floor(arg2)
			if in_layers < 1 then
				error("Must have at least one map layer.")
			end
			layers_per_map = in_layers
			i = i + 1
			
		elseif action == "zoom" then -- TODO: does this mess up on systems with highdpi?
			zoom = tonumber(arguments[i + 1])
			if not zoom or zoom <= 0.0 then
				error("zoom value needs to be a number > 0.")
			end
			i = i + 1
			
		else
			error("Unknown CLI argument: " .. tostring(action))
		end
		i = i + 1
	end
	
	-- Set starting size based on zoom and make window resizable
	love.window.setMode(800*zoom, 600*zoom, {resizable=true})
	
	map_cols = map_sq
	map_rows = map_sq
	
	tile_w = 16
	tile_h = 16
	
	n_tiles = map_cols*map_rows
	n_verts = n_tiles*4	
	
	-- Mock texture atlas
	atlas_cols, atlas_rows = 16, 16
	atlas_w, atlas_h = tile_w*atlas_cols, tile_h*atlas_rows
	do
		local i_data = love.image.newImageData(atlas_w, atlas_h)
		local temp_func = function(x, y, r, g, b, a)
			return x/atlas_w, x/atlas_w/2, y/atlas_h, 1.0
		end
		i_data:mapPixel(temp_func, 0, 0, atlas_w, atlas_h)
	
		atlas = love.graphics.newImage(i_data)
	end
	
	-- Mock quads representing a tileset
	quads = {}
	for y = 0, atlas_rows-1 do
		for x = 0, atlas_cols-1 do
			table.insert(quads, love.graphics.newQuad(x*tile_w, y*tile_h, tile_w, tile_h, atlas))
		end
	end
	
	-- Mock tilemap with random tiles
	map = {}
	pseudo = love.math.newRandomGenerator(27-12-2021)
	for i = 1, map_cols*map_rows do
		map[i] = pseudo:random(1, #quads)
	end
	
	-- Sets of batches+meshes.
	sets = {
		-- Static
		newBatchSet(atlas, layers_per_map, n_tiles, n_verts, vert_fmt, "static"),
		
		-- Dynamic
		newBatchSet(atlas, layers_per_map, n_tiles, n_verts, vert_fmt, "dynamic"),
		
		-- Stream
		newBatchSet(atlas, layers_per_map, n_tiles, n_verts, vert_fmt, "stream"),
	}
	
	-- Update SpriteBatches to reflect the map
	for i = 1, #sets do
		initBatch(sets[i])
	end
	
	love.graphics.setShader(mock_shader)
end


function love.keypressed(kc, sc)
	
	-- Change the active SpriteBatch + Mesh pair.
	if sc == "tab" then
		active_set = active_set + 1
		if active_set > #sets then
			active_set = 1
		end
	
	-- Toggle map rendering.
	elseif sc == "space" then
		render_map = not render_map
	
	-- Toggle whether a vertex is updated every frame or not.
	elseif sc == "return" then
		do_update_vertex = not do_update_vertex
		
	-- Toggle use of the shader when drawing the tilemap.
	elseif sc == "backspace" then
		use_shader = not use_shader
		if use_shader then
			love.graphics.setShader(mock_shader)
		else
			love.graphics.setShader()
		end
	
	-- Toggle VSync on/off.
	elseif sc == "lshift" then
		do_vsync = not do_vsync
		if do_vsync then
			love.window.setVSync(1)
		else
			love.window.setVSync(0)
		end
		
	-- Quit the application.	
	elseif sc == "escape" then
		love.event.quit()
	end
	
	-- Uncomment for debug commands
	--[[
	-- Dump all vertices in the active mesh to console
	if sc == "f12" then
		print("VERTEX DUMP:")
		for i = 1, n_verts do
			print(i, sets[active_set].mesh:getVertex(i))
		end
	end
	-- Render the atlas texture in love.draw()
	if sc == "f11" then
		show_atlas = not show_atlas
	end
	
	-- Print some details about the per-frame vertex write
	if sc == "f10" then
		dbg_print_vertex_update_info = not dbg_print_vertex_update_info
	end
	--]]
end


function love.mousemoved(x, y)
	-- Use Mouse X position to determine which vertex to update.
	local width, height = love.graphics.getDimensions()
	local x_scaled = math.floor((x / width) * n_verts)
	
	vertex_to_update = math.max(1, math.min(x_scaled + 1, n_verts))
end

--local t_vert_cont = {t_vert}
function love.update(dt)
	local layers = sets[active_set]
	
	time_set_vertex = love.timer.getTime()
	
	-- For each layer, update one vertex somewhere the attrib mesh.
	if do_update_vertex then
		for l = 1, #layers do
			local layer = layers[l]
			
			-- Make up some new attribute values in the range of (0-1)
			for i = 1, 4 do
				--t_vert[i] = 1.0
				t_vert[i] = pseudo:random()
			end
			
			if dbg_print_vertex_update_info then
				print("vertex_to_update", vertex_to_update, "l", l)
			end
			
			-- Update vertex
			layer.mesh:setVertex(vertex_to_update, t_vert)
			--layer.mesh:setVertices(t_vert_cont, vertex_to_update, 1)
			--layer.mesh:setVertices(vert_data, vertex_to_update, 1)
		end
	end
	
	time_set_vertex = love.timer.getTime() - time_set_vertex
end


function love.resize()
	local window_w = love.graphics.getDimensions()
	zoom = math.max(0.1, window_w / 800)
end


function love.draw()

	local layers = sets[active_set]
	
	time_draw_batch = love.timer.getTime()
	
	-- Mock map
	if render_map then
		love.graphics.push()
		
		local w_w, w_h = love.graphics.getDimensions()
		local m_w, m_h = map_cols * tile_w, map_rows * tile_h
		local scale = w_w / m_w
		
		for l = 1, #layers do
			love.graphics.draw(layers[l].batch, 0, 0, 0, scale)
		end
		
		love.graphics.pop()
	end
	
	love.graphics.scale(zoom, zoom)
	
	time_draw_batch = love.timer.getTime() - time_draw_batch
	
	-- Temporarily disable shader
	local banked_shader = love.graphics.getShader()
	love.graphics.setShader()
	
	-- Make stats a bit easier to read
	love.graphics.setColor(0, 0, 0, 0.75)
	love.graphics.rectangle("fill", 0, 0, 800, 164)
	love.graphics.setColor(1,1,1,1)
	
	-- (Debug: check the atlas)
	if show_atlas then
		love.graphics.setColor(0,0,0,1)
		love.graphics.rectangle("fill", 320-4, 320-4, atlas_w+8, atlas_h+8)
		love.graphics.setColor(1,1,1,1)
		love.graphics.draw(atlas, 320, 320)
	end
	
	-- Demo State / Controls
	love.graphics.print("Press Esc to Quit!", 16, 0)
	love.graphics.print("Mesh vertex to update (mouse X): " .. vertex_to_update, 16, 16)
	love.graphics.print("mesh set / usage (tab): " .. layers[1].usage, 16, 32)
	love.graphics.print("render_map (space): " .. tostring(render_map), 16, 48)
	love.graphics.print("update_vertex (return): " .. tostring(do_update_vertex), 16, 64)
	love.graphics.print("use_shader (backspace): " .. tostring(use_shader), 16, 80)
	love.graphics.print("VSync (left shift): " .. tostring(do_vsync), 16, 96)
	
	-- Stats
	love.graphics.print("n_verts: " .. n_verts, 16, 128)
	love.graphics.print("JIT enabled: " .. tostring(jit_enabled), 16, 144)

	-- Timing info
	love.graphics.print("FPS: " .. love.timer.getFPS(), 400, 16)
	love.graphics.print("AverageDelta: " .. love.timer.getAverageDelta(), 400, 32)
	
	love.graphics.print("Time to set vertex (*2^16): " .. time_set_vertex * (2^16), 400, 48)
	love.graphics.print("Time to draw batch: " .. time_draw_batch, 400, 64)
	
	-- More map state
	love.graphics.print("Layers per map: " .. layers_per_map, 400, 96)

	love.graphics.setShader(banked_shader)
end

